/***************************************************************************
    copyright            : (C) 2000 by Marko Grnroos
    email                : magi@iki.fi
 ***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************
 *
 **/

#include <magic/object.h>
#include <freenet.h>
#include <trainer.h>
#include <patternset.h>
#include <equalization.h>
#include "auditing.h"

impl_dynamic (AuditingStrategy, {Object});
impl_dynamic (AverageDeltaPrediction, {AuditingStrategy});
impl_dynamic (ZeroDeltaPrediction, {AuditingStrategy});
impl_dynamic (AbsoluteNeuralPrediction, {AuditingStrategy});
impl_dynamic (SingleNeuralPrediction, {AbsoluteNeuralPrediction});
impl_dynamic (DeltaNeuralPrediction, {AuditingStrategy});
impl_dynamic (PreviousYear, {AuditingStrategy});
impl_dynamic (PreviousYearsAvg, {AuditingStrategy});



/////////////////////////////////////////////////////////////////////////////////////////////////////
//   _             | o     o             -----                 ----                    |           //
//  / \            |    |      _           |    ___   ____  |  |   )  ___   ____       |  |   ____ //
// /   \ |   |  ---| | -+- | |/ \   ___    |   /   ) (     -+- |---  /   ) (     |   | | -+- (     //
// |---| |   | (   | |  |  | |   | (   \   |   |---   \__   |  | \   |---   \__  |   | |  |   \__  //
// |   |  \__!  ---| |   \ | |   |  ---/   |    \__  ____)   \ |  \   \__  ____)  \__! |   \ ____) //
//                                  __/                                                            //
/////////////////////////////////////////////////////////////////////////////////////////////////////

void AuditingTestResults::make (int variables) {
	averageError = 0.0;
	MSE = 0.0;
	RMSE = 0.0;
	averageErrors.make(variables);
	MSEs.make(variables);
	RMSEs.make(variables);
	for (int i=0; i<MSEs.size; i++) {
		averageErrors[i] = 0.0;
		MSEs[i] = 0.0;
		RMSEs[i] = 0.0;
	}
	mValues = 0;
}

void AuditingTestResults::addValue (int variable, double value) {
	averageError += value;
	MSE += value*value;
	averageErrors[variable] += value;
	MSEs[variable] += value*value;

	if (variable==0)
		mValues++;
}

void AuditingTestResults::calculate () {
	averageError /= mValues * MSEs.size;
	MSE /= mValues * MSEs.size;
	RMSE = sqrt (MSE);
	for (int i=0; i<MSEs.size; i++) {
		averageErrors[i] /= mValues;
		MSEs[i] /= mValues;
		RMSEs[i] = sqrt (MSEs[i]);
	}
}

void AuditingTestResults::operator>> (OStream& out) const {
	printf ("Average error = %f\n", averageError);
	printf ("Mean Squared Error = %f\n", MSE);
	printf ("Root of Mean Squared Error = %f\n", RMSE);
	printf ("Variable\tavgErr\t\tMSE\t\tRMSE\n");
	for (int i=0; i<MSEs.size; i++)
		printf ("%3d\t\t%#5.10g\t%#5.10g\t%5.10g\n",
				i+1, averageErrors[i], MSEs[i], RMSEs[i]);
}


/////////////////////////////////////////////////////////////////////////////////////
//   _             | o     o              ----                                     //
//  / \            |    |      _         (      |       ___   |   ___              //
// /   \ |   |  ---| | -+- | |/ \   ___   ---  -+- |/\  ___| -+- /   )  ___  \   | //
// |---| |   | (   | |  |  | |   | (   \     )  |  |   (   |  |  |---  (   \  \  | //
// |   |  \__!  ---| |   \ | |   |  ---/ ___/    \ |    \__|   \  \__   ---/   \_/ //
//                                  __/                                 __/   \_/  //
/////////////////////////////////////////////////////////////////////////////////////

/*virtual*/ void AuditingStrategy::make (const StringMap& params) {
	mInputMonths	= int(params["inputMonths"]);
}

/*virtual*/ AuditingTestResults* AuditingStrategy::test (const Matrix& testdata, int startmonth) const {
	// Let the strategy do the prediction
	Matrix testres = predict (testdata, startmonth);

	////////////////////////////////////////
	// Collect error statistics for each pattern

	// Create result storage object
	AuditingTestResults* result = new AuditingTestResults ();
	result->make (testdata.cols);

	// Iterate through test patterns. Skip the mInputMonths, since they are
	// not actual test data points.
	for (int r=mInputMonths; r<testdata.rows; r++) {
		// Remove year from the month number, add index, transform to 0-based, and wrap around 12
		int month = ((startmonth % 100)+r-1)%12;
		
		// Calculate error for one test pattern
		for (int c=0; c<testdata.cols; c++) {
			// Calculate error between the desired and predicted value
			double error = fabs (testdata.get(r,c) - testres.get(month,c));
			//printf ("row=%d, col=%d, error=%f\n", r, c, error);

			// Add one sample to the statistics calculator
			result->addValue (c, error);
		}
	}
	printf ("testing done\n");

	// Finalize the statistics calculation
	result->calculate ();
	return result;
}

/*virtual*/ void AuditingStrategy::testCurve (const Matrix& testdata, int startmonth, const String& filename) const {
}

/*static*/ Matrix AuditingStrategy::rowDeltas (const Matrix& matrix) return deltas {
	deltas.make (matrix.rows-1, matrix.cols);
	for (int r=0; r<deltas.rows; r++)
		for (int c=0; c<deltas.cols; c++)
			deltas.get(r,c) = matrix.get(r+1,c) - matrix.get(r,c);
}



///////////////////////////////////////////////////////////////////////////////
//    _                                     ___         |           ----     //
//   / \         ___       ___         ___  |  \   ___  |  |   ___  |   )    //
//  /   \ |   | /   ) |/\  ___|  ___  /   ) |   | /   ) | -+-  ___| |---     //
//  |---|  \ /  |---  |   (   | (   \ |---  |   | |---  |  |  (   | |        //
//  |   |   V    \__  |    \__|  ---/  \__  |__/   \__  |   \  \__| |     O  //
//                               __/                                         //
///////////////////////////////////////////////////////////////////////////////

AverageDeltaPrediction::AverageDeltaPrediction (const StringMap& params) : AuditingStrategy ("AverageDelta") {
}

/*virtual*/ void AverageDeltaPrediction::train (const Matrix& traindata, int startmonth) {
	ASSERT ((int(fabs(traindata.get(0,0)-traindata.get(1,0)+0.001))%100) != 1);
	printf ("Training data:");
	sout << traindata;
	
	// Calculate deltas
	Matrix deltas = rowDeltas (traindata);
	
	// Initialize average matrix
	mMonthAvg.make (12, deltas.cols);
	mMonthCnt.make (12);
	for (int r=0; r<mMonthAvg.rows; r++) {
		for (int c=0; c<mMonthAvg.cols; c++)
			mMonthAvg.get(r,c) = 0.0;
		mMonthCnt[r] = 0;
	}
	
	// Calculate average deltas for each month
	int startmonthi = startmonth%100;
	for (int r=0; r<deltas.rows; r++) {
		int month = (startmonthi+r-1+1+12)%12;
		for (int c=0; c<deltas.cols; c++)
			mMonthAvg.get(month,c) = (mMonthAvg.get(month,c)*mMonthCnt[month]
									  + deltas.get(r,c))/ (mMonthCnt[month]+1);
		mMonthCnt[month]++;
	}
}

/*virtual*/ Matrix AverageDeltaPrediction::predict (const Matrix& testdata, int startmonth) const return result {
	printf ("Test data:");
	sout << testdata;
	
	int startmonthi = startmonth%100;
	result.make (testdata.rows-inputMonths(), testdata.cols);
	for (int r=0; r<result.rows; r++) {
		for (int c=0; c<testdata.cols; c++) {
			int month = (startmonthi+inputMonths()+r-1)%12;
			result.get(r,c) = testdata.get(r+inputMonths()-1,c) + mMonthAvg.get(month, c);
		}
	}
}



//////////////////////////////////////////////////////////////////////////////
//           -----                ___         |           ----              //
//              /   ___           |  \   ___  |  |   ___  |   )             //
//             /   /   ) |/\  __  |   | /   ) | -+-  ___| |---              //
//            /    |---  |   /  \ |   | |---  |  |  (   | |                 //
//           /____  \__  |   \__/ |__/   \__  |   \  \__| |     O           //
//////////////////////////////////////////////////////////////////////////////

ZeroDeltaPrediction::ZeroDeltaPrediction (const StringMap& params) : AuditingStrategy ("ZeroDelta") {
}

/*virtual*/ void ZeroDeltaPrediction::train (const Matrix& traindata, int startmonth) {
}

/*virtual*/ Matrix ZeroDeltaPrediction::predict (const Matrix& testdata, int startmonth) const return result {
	ASSERT (mInputMonths>=1);

	int startmonthi = startmonth%100;
	result.make (testdata.rows-mInputMonths, testdata.cols);
	for (int r=0; r<result.rows; r++)
		for (int c=0; c<testdata.cols; c++)
			result.get(r,c) = testdata.get(r+mInputMonths-1,c);
}



////////////////////////////////////////////////////////////////////////////////////
//   _                    |                 |   |                       | ----    //
//  / \  |      ____      |        |   ___  |\  |  ___             ___  | |   )   //
// /   \ |---  (      __  | |   | -+- /   ) | \ | /   ) |   | |/\  ___| | |---    //
// |---| |   )  \__  /  \ | |   |  |  |---  |  \| |---  |   | |   (   | | |       //
// |   | |__/  ____) \__/ |  \__!   \  \__  |   |  \__   \__! |    \__| | |     O //
////////////////////////////////////////////////////////////////////////////////////

AbsoluteNeuralPrediction::~AbsoluteNeuralPrediction	() {
	delete mpNetwork;
}

void AbsoluteNeuralPrediction::make (const StringMap& params) {
	mUseAllOutputs		= int(params["AbsoluteNeuralPrediction.useAllOutputs"]);
	mUseAllInputs		= int(params["AbsoluteNeuralPrediction.useAllInputs"]);
	ASSERT (mUseAllInputs || (!mUseAllInputs && !mUseAllOutputs));

	mGlobalEqualization	= int(params["AbsoluteNeuralPrediction.globalEqualization"]);
	mHiddenTopology		= params["AbsoluteNeuralPrediction.hidden"];
	mpNetwork			= NULL;

	mParams = params;

	AuditingStrategy::make (params);
}

PatternSet* AbsoluteNeuralPrediction::makeSet (const Matrix& data, int startmonth) const {
	TRACE2 ("Making pattern set from = %d rows, %d cols", data.rows, data.cols);
	
	int startmonthi = (startmonth % 100)-1;
	int inputs = data.cols;
	
	// Input patterns have inputs for each month (12) + inputmonths*attributes.
	PatternSet* result = new PatternSet (data.rows-inputMonths(),
										 12+inputMonths()*(mUseAllInputs? inputs:1),
										 mUseAllOutputs? inputs:1);

	// Copy the monthly attributes to the pattern set
	for (int p=0; p<result->patterns; p++) {
		// Zero the input month indicator flags
		for (int m=0; m<12; m++)
			result->set_input (p, m, 0);

		// Set input month indicators as 1
		for (int im=0; im<mInputMonths; im++)
			result->set_input (p, (startmonthi+p+im+12)%12, 1);
				
		// Copy input month values to inputs
		if (mUseAllInputs)
			for (int j=0; j<inputs; j++)
				for (int prm=0; prm<mInputMonths; prm++)
					result->set_input (p, 12+inputs*prm+j, data.get(p+prm, j));
		else {
			// Only one input variable
			for (int prm=0; prm<mInputMonths; prm++)
				result->set_input (p, 12+prm, data.get(p+prm, mVariable));
		}
		
		// Copy current month values to outputs
		if (mUseAllOutputs)
			for (int j=0; j<inputs; j++)
				result->set_output (p, j, data.get(p+mInputMonths, j));
		else
			// Only one output variable
			result->set_output (p, 0, data.get(p+mInputMonths, mVariable));
	}
	return result;
}

/*virtual*/ void AbsoluteNeuralPrediction::train (const Matrix& traindata, int startmonth) {
	TRACE2 ("Training data = %d rows, %d cols", traindata.rows, traindata.cols);
	
	////////////////////////////////////////////////////////////////////////////////
	// Create network
	
	if (!mpNetwork) {
		mpNetwork = new ANNetwork;
		mpNetwork->make (format("%d%s%d", inputVariables(traindata.cols),
								(CONSTR) mHiddenTopology,
								outputVariables(traindata.cols)));
		mpNetwork->connectFullFfw (false);
	}

	// Initialize the network
	mpNetwork->init (0.5);

	////////////////////////////////////////////////////////////////////////////////
	// Prepare data

	// Equalize data
	Matrix equalized = traindata;
	if (!mpNetwork->getEqualizer()) {
		MatrixEqualizer* mequalizer = new MatrixEqualizer (new MinmaxEq(0.0, 1.0)); // new HistogramEq (100000, 0.0, 1.0)
		mequalizer->analyze (equalized, mGlobalEqualization);
		mpNetwork->setEqualizer (mequalizer);
	}
	dynamic_cast<MatrixEqualizer*>(mpNetwork->getEqualizer())->equalize (equalized);

	// Create pattern set
	PatternSet* trainset = makeSet (equalized, startmonth);
	fprintf (stderr, "Training data has %d patterns with %d inputs and %d outputs\n",
			 trainset->patterns, trainset->inputs, trainset->outputs);

	PatternSet* trainset2 = makeSet (traindata, startmonth);
	delete trainset2;
	
	////////////////////////////////////////////////////////////////////////////////
	// Train

	// Create trainer and set parameters
	RPropTrainer trainer;
	trainer.init (mParams);
	trainer.setTerminator (mParams["terminator"]);

	// Set observer to display current training cycle
	if (rpObserver)
		trainer.setObserver (rpObserver);

	// Train
	double trainmse = trainer.train (*mpNetwork, *trainset, int(mParams["maxCycles"]),
									 NULL, int(mParams["validationInterval"]));

	fprintf (stderr, "Trained for %d cycles, MSE=%f\n", trainer.totalCycles(), trainmse);
	delete trainset;
}

/*virtual*/ Matrix AbsoluteNeuralPrediction::predict (const Matrix& testdata, int startmonth) const return result {
	TRACE2 ("Test data = %d rows, %d cols", testdata.rows, testdata.cols);
	
	// Equalize
	Matrix equalized = testdata;
	dynamic_cast<MatrixEqualizer*>(mpNetwork->getEqualizer())->equalize (equalized);
	
	// Create pattern set and use it
	PatternSet* testset = makeSet (equalized, startmonth);

	////////////////////////////////////////////////////////////////////////////////
	// Test the network

	// Create a matrix where to place the results
	result.make (testset->patterns, mUseAllOutputs? testset->outputs:1);

	// Test one month at a time
	for (int p=0; p<testset->patterns; p++) {
		Vector v = mpNetwork->testPattern (*testset, p);
		for (int i=0; i<v.size; i++)
			result.get(p,i) = v[i];
	}
	delete testset;

	// Unequalize the results to get money values again
	dynamic_cast<MatrixEqualizer*>(mpNetwork->getEqualizer())->unequalize (result);
}

void AbsoluteNeuralPrediction::setNetwork (ANNetwork* net) {
	delete mpNetwork;
	mpNetwork = net;
}



///////////////////////////////////////////////////////////////////////////////
//      ---- o             |       |   |                       | ----        //
//     (         _         |  ___  |\  |  ___             ___  | |   )       //
//      ---  | |/ \   ___  | /   ) | \ | /   ) |   | |/\  ___| | |---        //
//         ) | |   | (   \ | |---  |  \| |---  |   | |   (   | | |           //
//     ___/  | |   |  ---/ |  \__  |   |  \__   \__! |    \__| | |     O     //
//                    __/                                                    //
///////////////////////////////////////////////////////////////////////////////

void SingleNeuralPrediction::make (const StringMap& params) {
	AbsoluteNeuralPrediction::make (params);
}

/*virtual*/ void SingleNeuralPrediction::train (const Matrix& traindata, int startmonth) {
	for (int v=0; v<mNetworks.size; v++)
		mNetworks.cut (v);
	mNetworks.empty ();
	mNetworks.make (traindata.cols);

	// Train each variable separately
	sout.autoflush ();
	for (int v=0; v<traindata.cols; v++) {
		sout.printf ("Training variable %d...\n", v);
		mVariable = v;
		AbsoluteNeuralPrediction::train (traindata, startmonth);
		mNetworks.put (mpNetwork, v);
		mpNetwork=NULL;
	}
	sout.printf("\n");
}

/*virtual*/ Matrix SingleNeuralPrediction::predict (const Matrix& testdata, int startmonth) const return result {
	SingleNeuralPrediction* ncthis = const_cast<SingleNeuralPrediction*>(this);

	// For each variable
	for (int v=0; v<9; v++) {
		ncthis->mVariable = v;

		// Use a small kludge to set the current network
		ncthis->mpNetwork = const_cast<ANNetwork*>(&mNetworks[v]);

		// Let the baseclass do the prediction for one variable
		Matrix oneVariable = AbsoluteNeuralPrediction::predict (testdata, startmonth);

		// Create the result matrix when we get the number of rows.
		if (result.rows==0)
			result.make (oneVariable.rows, 9);
		
		// Copy the one variable to the total result
		for (int r=0; r<result.rows; r++)
			result.get (r, v) = oneVariable.get (r, 0);
	}
	ncthis->mpNetwork = NULL;
}



///////////////////////////////////////////////////////////////////////////////
//       ___         |           |   |                       | ----          //
//       |  \   ___  |  |   ___  |\  |  ___             ___  | |   )         //
//       |   | /   ) | -+-  ___| | \ | /   ) |   | |/\  ___| | |---          //
//       |   | |---  |  |  (   | |  \| |---  |   | |   (   | | |             //
//       |__/   \__  |   \  \__| |   |  \__   \__! |    \__| | |     O       //
///////////////////////////////////////////////////////////////////////////////

DeltaNeuralPrediction::DeltaNeuralPrediction (const StringMap& params) : AuditingStrategy ("DeltaNeural") {
}

/*virtual*/ void DeltaNeuralPrediction::train (const Matrix& traindata, int startmonth) {
}

/*virtual*/ Matrix DeltaNeuralPrediction::predict (const Matrix& testdata, int startmonth) const {
}



//////////////////////////////////////////////////////////////////////////////
//                                                                          //
//         ----                  o                                          //
//         |   )      ___                      ____  ___   ___              //
//         |---  |/\ /   ) |   | |  __  |   | (     /   )  ___| |/\         //
//         |     |   |---   \ /  | /  \ |   |  \__  |---  (   | |           //
//         |     |    \__    V   | \__/  \__! ____)  \__   \__| |           //
//                                                                          //
//////////////////////////////////////////////////////////////////////////////

PreviousYear::PreviousYear (const StringMap& params) : AuditingStrategy ("PreviousYear") {
}

/*virtual*/ void PreviousYear::train (const Matrix& traindata, int startmonth) {
	ASSERT (traindata.rows >= 12);
	
	// One year
	mData.make (12, traindata.cols);
	for (int r=0; r<mData.rows; r++) {
		for (int c=0; c<mData.cols; c++) {
			mData.get(r,c) = traindata.get (traindata.rows-12+r,c);
			printf ("%f ", mData.get(r,c));
		}
		printf ("\n");
	}
}

/*virtual*/ Matrix PreviousYear::predict (const Matrix& testdata, int startmonth) const  return result {
	ASSERT (mData.rows>0);
	ASSERT (testdata.rows == 12);

	result = mData;
	//result.make (mData.rows, mData.cols);
	//for (int r=0; r<mData.rows; r++)
	//	for (int c=0; c<mData.cols; c++)
	//		result.get (r,c) = mData.get (r,c);
}



//////////////////////////////////////////////////////////////////////////////////////
//                                                                                  //
// ----                  o                                          _               //
// |   )      ___                      ____  ___   ___       ____  / \              //
// |---  |/\ /   ) |   | |  __  |   | (     /   )  ___| |/\ (     /   \ |   |  ___  //
// |     |   |---   \ /  | /  \ |   |  \__  |---  (   | |    \__  |---|  \ /  (   \ //
// |     |    \__    V   | \__/  \__! ____)  \__   \__| |   ____) |   |   V    ---/ //
//                                                                             __/  //
//////////////////////////////////////////////////////////////////////////////////////

PreviousYearsAvg::PreviousYearsAvg (const StringMap& params) : AuditingStrategy ("PreviousYearsAvg") {
}

/*virtual*/ void PreviousYearsAvg::train (const Matrix& traindata, int startmonth) {
	ASSERT (traindata.rows >= 24);
	
	// One year
	mMonthlyAvg.make (12, traindata.cols);
	for (int r=0; r<mMonthlyAvg.rows; r++)
		for (int c=0; c<mMonthlyAvg.cols; c++) {
			// Calculate average of previous years
			float sum = 0.0;
			int year = 0;
			for (; traindata.rows-12+r-year*12 >= 0; year++)
				sum += traindata.get (traindata.rows-12+r-year*12, c);
			mMonthlyAvg.get(r,c) = sum/float(year);
		}
}

/*virtual*/ Matrix PreviousYearsAvg::predict (const Matrix& testdata, int startmonth) const return result {
	ASSERT (mMonthlyAvg.rows>0);
	ASSERT (testdata.rows == 12);
	
	result = mMonthlyAvg;
}



///////////////////////////////////////////////////////////////////////////////////////////////////
//   _             | o     o              ----                                     |     o       //
//  / \            |    |      _         (      |       ___   |   ___              |       |     //
// /   \ |   |  ---| | -+- | |/ \   ___   ---  -+- |/\  ___| -+- /   )  ___  \   | |     | |---  //
// |---| |   | (   | |  |  | |   | (   \     )  |  |   (   |  |  |---  (   \  \  | |     | |   ) //
// |   |  \__!  ---| |   \ | |   |  ---/ ___/    \ |    \__|   \  \__   ---/   \_/ |____ | |__/  //
//                                  __/                                 __/   \_/                //
///////////////////////////////////////////////////////////////////////////////////////////////////

AuditingStrategyLib AuditingStrategyLib::sInstance;

AuditingStrategyLib::AuditingStrategyLib () {
	registerCls ("AverageDeltaPrediction");
	registerCls ("ZeroDeltaPrediction");
	registerCls ("AbsoluteNeuralPrediction");
	registerCls ("SingleNeuralPrediction");
	registerCls ("PreviousYear");
	registerCls ("PreviousYearsAvg");
	//registerCls ("DeltaNeuralPrediction");
}

AuditingStrategy* AuditingStrategyLib::create (int i) const {
	return dynamic_cast<AuditingStrategy*> (dyncreate (mClassNames[i]));
}


