#include <magic/Math.h> 
#include "annetwork.h" 
#include "patternset.h"
#include "termination.h"
#include "freenet.h"


///////////////////////////////////////////////////////////////////////////////
//                                                                           //
//                _   |   | |   |                           |                //
//               / \  |\  | |\  |  ___   |                  |                //
//              /   \ | \ | | \ | /   ) -+- \    /  __  |/\ | /              //
//              |---| |  \| |  \| |---   |   \\//  /  \ |   |/               //
//              |   | |   | |   |  \__    \   VV   \__/ |   | \              //
//                                                                           //
///////////////////////////////////////////////////////////////////////////////

AnyNetwork::AnyNetwork () {
}

double AnyNetwork::test (const PatternSet& set) const {
	ASSERT (set.patterns>0);
	
	double errorSum = 0.0; // Sum of squared errors (SSE)
	for (int p=0; p<set.patterns; p++) {
		Vector res = testPattern (set, p);
		ASSERT (res.size == set.outputs);
		
		for (int j=0; j<res.size; j++)
			errorSum += sqr (res[j] - set.output (p, j));
	}
	
	return errorSum/(set.patterns*set.outputs); // Mean of squared errors (MSE)
}

ClassifResults* AnyNetwork::testClassify (const PatternSet& set) const {
	ASSERT (set.patterns>0);

	ClassifResults* result = new ClassifResults;

	// Determine the number of classes (not so trivial)
	int classes = (set.outputs>1)? set.outputs : 2;
				   
	// Initialize the class hit counter
	result->classcnts.make (classes);
	for (int i=0; i<classes; i++)
		result->classcnts[i] = 0;

	// Initialize the class instance counter
	result->classSizes.make (classes);
	for (int i=0; i<classes; i++)
		result->classSizes[i] = 0;
	
	// Classify each pattern in the set
	int failures=0;
	double errorSum = 0.0; // Sum of squared errors (SSE)
	for (int p=0; p<set.patterns; p++) {
		// Find the correct class 
		int correctClass = set.getClass (p);
		
		// Determine success
		bool success=false;

		Vector res = testPattern (set, p);

		// Record the SSE
		for (int j=0; j<res.size; j++)
			errorSum += sqr (res[j] - set.output (p, j));

		if (set.outputs==1) {
			if (correctClass == int(res[0]+0.5))
				success = true;
		} else {
			int highestClass = maxIndex (res);
			if (highestClass == correctClass)
				success = true;
		}
		
		// Record the success
		if (!success) {
			failures++;
			
			// For the particular class
			result->classcnts[correctClass]++;
		}

		// And increment the number of instances for this particular class
		result->classSizes[correctClass]++;
	}
	
	// Return the mean
	result->mse = errorSum/(set.patterns*set.outputs); // Mean of squared errors (MSE)
	result->failures = failures;

	return result;
}

void AnyNetwork::copyFreeNet (const ANNetwork& fnet, bool onlyWeights) {
	copy (static_cast<const AnyNetwork&> (fnet));
}

void AnyNetwork::copy (const AnyNetwork& fnet, bool onlyWeights) {
}



