#include <magic/Class.h>
#include "neuron.h"
#include "connection.h"

impl_dynamic (Connection, {Object});


///////////////////////////////////////////////////////////////////////////////
//             ___                                   o                       //
//            /   \        _     _    ___   ___   |           _              //
//            |      __  |/ \  |/ \  /   ) |   \ -+- |  __  |/ \             //
//            |     /  \ |   | |   | |---  |      |  | /  \ |   |            //
//            \___/ \__/ |   | |   |  \__   \__/   \ | \__/ |   |            //
///////////////////////////////////////////////////////////////////////////////

Connection::Connection (const Neuron* source, const Neuron* target, double weight) {
	mpSource = const_cast<Neuron*>(source);
	mpTarget = const_cast<Neuron*>(target);
	mWeight  = weight;
}

Connection::~Connection () {
	if (mpSource)
		mpSource->disconnectTo (*this);
	if (mpTarget)
		mpTarget->disconnectFrom (*this);
}

void Connection::init (double r) {
	mWeight = 2*r*frnd()-r;
}

double Connection::transfer () {
	return mWeight * mpSource->output();
}

void Connection::copy (const Connection& other) {
	mWeight = other.weight();

#warning "TODO: Copying connections not implemented"
	mpSource = NULL;
	mpTarget = NULL;
}


void Connection::check (int netSize) const {
	ASSERT (mpSource!=NULL);
	ASSERT (mpTarget!=NULL);
}

void Connection::writeXML (OStream& out) const {
	sout.printf ("\t\t<%s>", (CONSTR) getclassname());
	if (mpSource)
		sout.printf ("<SOURCEID>%d</SOURCEID>", mpSource->id());
	sout.printf ("<WEIGHT>%f</WEIGHT>", mWeight);
	sout.printf ("</%s>\n", (CONSTR) getclassname());
}
