/***************************************************************************
    copyright            : (C) 2000 by Marko Grnroos
    email                : magi@iki.fi
 ***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************
 *
 **/

#include <fstream.h>

#include <magic/object.h>

#include "dataformat.h"
#include "dataformats.h"



//////////////////////////////////////////////////////////////////////////////
//    ___                   -----                          |     o          //
//    |  \   ___   |   ___  |                     ___   |  |       |        //
//    |   |  ___| -+-  ___| |---   __  |/\ |/|/|  ___| -+- |     | |---     //
//    |   | (   |  |  (   | |     /  \ |   | | | (   |  |  |     | |   )    //
//    |__/   \__|   \  \__| |     \__/ |   | | |  \__|   \ |____ | |__/     //
//////////////////////////////////////////////////////////////////////////////

void DataFormatLib::load (const String& filename, PatternSet& set)  throw (file_not_found, invalid_format, assertion_failed) {
	ASSERTWITH (!isempty(filename), "Filename required (was empty)");

	// Open the file
	istream* in = &cin;
	if (filename!="-" && !filename.isempty()) {
		in = new ifstream ((CONSTR) filename);
		if (!*in)
			throw file_not_found (format ("Pattern set file '%s' not found", (CONSTR) filename));
	}

	load (*in, set, filename);

	if (in != &cin)
		delete in;
}

void DataFormatLib::load (istream& in, PatternSet& set, const String& filetype=".raw") throw (invalid_format, assertion_failed) {
	DataFormat* handler = create (filetype);
	try {
		handler->load (in, set);
	} catch (...) {
		delete handler;
		throw; // Rethrow
	}
	delete handler;
}

void DataFormatLib::save (const String& filename, const PatternSet& set) throw (invalid_filename) {
	FILE* out=stdout;
	if (filename!="-" && !filename.isempty())
		out = fopen (filename, "w");

	ASSERTWITH (out, format ("Pattern save file '%s' couldn't be opened",
							 (CONSTR) filename));

	DataFormat* handler = create (filename);

	try {
		handler->save (out, set);
	} catch (must_overload e) {
		delete handler;
		throw invalid_filename (format ("Save-to-file operation not supported for file type\n%s", (CONSTR) e.what()));
	}
	delete handler;

	if (out != stdout)
		fclose (out);
}

DataFormat* DataFormatLib::create (const String& filename) {
	// Parse the contents according to the file name extension
	if (filename.len>4 && filename.right(4)==".pat")
		return new SNNSDataFormat ();
	else if (filename.len>3 && filename.right(3)==".dt")
		return new Proben1DataFormat ();
	else
		return new RawDataFormat ();
}

